{*******************************************************}
{                                                       }
{              XmlSqlDb Component Library               }
{       Pal Sitkei (sitkei@chello.hu, 2004.11.13)       }
{                                                       }
{*******************************************************}

unit XmlSqlDb;

interface

uses
  Classes, Consts, Db;

{$DEFINE FULL}

type                        
  TLogFile = class(TStringList)
  private
    { Private declarations }
    FPath, FFileName, FError: String;
    FMaxLine: Integer;
    procedure SetError(AValue: String);
    procedure SetPath(AValue: String);
  public
    { Public declarations }
    function Add(const S: string): Integer; override;
    property RunError: String read FError write SetError;
    property Path: String read FPath write SetPath;
    property FileName: String read FFileName write FFileName;
    property MaxLine: Integer read FMaxLine write FMaxLine;
  end;

  tSetOfChar = set of Char;

  tAwkline = class(tStringList)
  protected
    setGarbagge: tSetOfChar;
    function IsGarbagge(C: Char): Boolean;
  public
    constructor Create(S: String; setCh: tSetOfChar);
    function GetString(Index: Integer): String;
  end;

  tSqlParser = class(tStringList)
  private
    FDelimiter, FOperator: tSetOfChar;
    FParseStr: String;
    FQueryList: TList;
    procedure ConnectStrings;
    procedure ConnectEnum;
    procedure ConnectOrderBy;
    procedure Parse;
    function PushSpaces: Boolean;
    procedure RemoveComments;
    procedure RemoveOuterBrackets;
    procedure SetParseStr(aValue: String);
    procedure SplitParse;
    procedure Upper;
  public
    constructor Create;
    destructor Destroy; override;
    function GetQuery(Index: Integer): tStringList;
    property Delimiter: tSetOfChar read FDelimiter write FDelimiter;
    property Operator: tSetOfChar read FOperator write FOperator;
    property ParseStr: String read FParseStr write SetParseStr;
    property QueryList: tList read FQueryList write FQueryList;
  end;

  PRecInfo = ^TRecInfo;
  TRecInfo = packed record
    BookMark: Longint;
    UpdateStatus: TUpdateStatus;
    BookmarkFlag: TBookmarkFlag;
  end;

  tEval = class(tComponent)
  private
    WhatIndex: Integer;
  public
    Expression, Fields, Types, Recor: tStringList;
    constructor Create(AOwner: tComponent); override;
    destructor Destroy; override;
    function GetResult: Boolean;
    procedure Setup;
  end;

  TArrInt = array [0..10000] of integer;
  PInteger = ^TArrInt;
  PLongInt = ^LongInt;
  TOrderFieldType = (oftBoolean, oftDatetime,oftFloat,oftInteger,oftString);

  TStrGrid = class(TStringList)
  private
    FRow, FCol : Integer;
    function  GetCells(ACol, ARow: Integer): string;
    function  GetRObject(ARow: Integer): TObject;
    procedure PutRObject(ARow: Integer; const AObject: TObject);
    procedure SetCells(ACol, ARow: Integer; const Value: string);
  protected
    FDataBase, FFileName, FSql, FBakFile: String;
    FParser: tSqlParser;
    FFields, FFieldU, FTypes: tStringList;
    FLastId: Integer;
    procedure AnalyseSql;
    procedure CopyFromResult(aRes: TStrGrid);
    function GetOriginalName(aName: String): String;
    procedure MakeEmptyDataSet(aMessage: String);
    function MakeTableHeader: String;
    function ReadTableData(Line: String): Boolean;
    procedure SetOrderFieldType;
    procedure SetSql(aValue: String);
    function StripComas(aStr: String): String;
    procedure XmlCreate;
    procedure XmlDelete;
    procedure XmlDrop;
    procedure XmlInsert;
    procedure XmlSelect;
    procedure XmlUpdate;
  public
    constructor Create;
    destructor Destroy; override;
    procedure DeleteRow( Index : integer );
    function  IndexOfObject( AObject : TObject ) : Integer; override;
    property  DataBase: string read FDataBase write FDataBase;
    property  Parser: tSqlParser read FParser write FParser;
    property  Sql: string read FDataBase write SetSql;
    property  Cells[ACol, ARow: Integer]: string read GetCells write SetCells; default;
    property  RowObjects[ARow: Integer]: TObject read GetRObject write PutRObject;
    property  RowCount : Integer read FRow write Frow;
    property  ColCount : Integer read FCol write Fcol;
  end;

  TXmlSqlDb = class(TDataSet)
  private
    FLogFile: TLogFile;
    FDataBase, FSql, FUndoDataBase, FUndoFileName, FUndoBakFile: String;
    FFldOffs       : PInteger;
    FData          : TStrGrid;
    FFldNames      : TStrings;
    FRecSize, FRecBufSize, FRecInfoOfs, FCurRec, FLastBookmark: Integer;
    function GetCanUndo: Boolean;
    function GetErrorStr: String;
    function GetLogText: String;
  protected
    function  AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function  GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function  GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function  IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function  GetRecordCount: Integer; override;
    function  GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;
  public
    constructor Create(aOwner: tComponent); override;
    destructor Destroy; override;
    function  GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    procedure Undo;
    property CanUndo: Boolean read GetCanUndo;
    property ErrorStr: String read GetErrorStr;
    property LogText: String read GetLogText;
  published
    property Sql: string read FSql write FSql;
    property DataBase: string read FDataBase write FDataBase;
    property Active;
  end;

procedure Register;

implementation

uses SysUtils, Forms;

var LogFile: TLogFile;
    OrderIndex: Integer;
    OFT: TOrderFieldType;

const MaxRecLength  = 4096;

procedure Register;
begin
  RegisterComponents('Samples', [TXmlSqlDb]);
end;

// TLogFile

procedure TLogFile.SetError(AValue: String);
begin
  FError:= AValue;
  Add(AValue);
end;

procedure TLogFile.SetPath(AValue: String);
begin
  FPath:= AValue;
  if FPath[Length(FPath)]<> '\' then FPath:= FPath+ '\';
end;

function TLogFile.Add(const S: string): Integer;
begin
  Insert(0, DateToStr(Now)+ ' '+ TimeToStr(Now)+ ' '+ S);
  if Count > FMaxLine then while Count > FMaxLine do Delete(0);
//  SaveToFile(FPath+ FFileName);
  Result:= Count- 1;
end;

// TAwkLine

constructor tAwkline.Create(S: String; setCh: tSetOfChar);
var i: Integer;
    Temp: String;
begin
  inherited Create;
  setGarbagge:= setCh;
  Temp:= '';
  for i:=1 to Length(S) do
    if not IsGarbagge(S[i]) then Temp:= Temp+ S[i] else
    begin
      Add(Temp);
      Temp:= '';
    end;
  if Temp<> '' then Add(Temp);
end;

function tAwkline.IsGarbagge(C: Char): Boolean;
begin
  Result:= C in setGarbagge;
end;

function tAwkline.GetString(Index: Integer): String;
begin
  Result:= Strings[Index];
end;

// TSqlParser

procedure tSqlParser.ConnectStrings;
var i, po: Integer;
    IsString: Boolean;
    Coll: String;
    Temp: tStringList;
begin
  Temp:= tStringList.Create;
  IsString:= False;
  for i:= 0 to Count-1 do
  begin
    if IsString then
    begin
      Coll:= Coll+ Strings[i];
      IsString:= (Pos('"', Strings[i]) = 0);
      if not IsString then Temp.Add(Coll)
                      else Coll:= Coll+ ' ';
    end else
    begin
      po:= Pos('"', Strings[i]);
      if (po<>0) and (Pos('"', Copy(Strings[i], po+ 1, Length(Strings[i])))= 0) then
      begin
        IsString:= True;
        Coll:= Strings[i]+ ' ';
      end else Temp.Add(Strings[i]);
    end;
  end;
  Clear;
  for i:= 0 to Temp.Count-1 do Add(Temp.Strings[i]);
  Temp.Free;
end;

procedure tSqlParser.ConnectEnum;
var i: Integer;
    IsComa: Boolean;
    Temp: tStringList;
    S1: String;
begin
  Temp:= tStringList.Create;
  IsComa:= False;
  for i:= 0 to Count-1 do
  begin
    S1:= Strings[i];
    if ((Strings[i] = ',')or(Pos(',', Strings[i])=Length(Strings[i])))then
    begin
      IsComa:= True;
      Temp.Strings[Temp.Count- 1]:= Temp.Strings[Temp.Count- 1]+ Strings[i];
    end else
    begin
      if IsComa then
      begin
        Temp.Strings[Temp.Count- 1]:= Temp.Strings[Temp.Count- 1]+ Strings[i];
        IsComa:= False;
      end else
        Temp.Add(Strings[i]);
    end;
  end;
  Clear;
  for i:= 0 to Temp.Count-1 do Add(Temp.Strings[i]);
  Temp.Free;
end;

procedure tSqlParser.ConnectOrderBy;
var i: Integer;
begin
  for i:= 0 to Count-1 do
    if ((Strings[i] = 'ORDER')and(Strings[i+ 1] = 'BY'))then
    begin
      Strings[i]:= 'ORDER BY';
      Delete(i+ 1);
      Break;
    end;
end;

procedure tSqlParser.Parse;
var i: Integer;
    Temp, CloseChar: String;
begin
  Clear;
  Temp:= '';
  CloseChar:= '';
  for i:=1 to Length(FparseStr) do
  begin
    if (CloseChar <> '') then
    begin
      if (FparseStr[i] = CloseChar) then
      begin
        CloseChar:= '';
        Temp:= Temp+ FparseStr[i];
      end else
      begin
        Temp:= Temp+ FparseStr[i];
      end;
    end else
    begin
      if (FparseStr[i] in FDelimiter) then
      begin
        if Temp>'' then Add(Temp);
        Temp:= '';
      end else
      begin
        Temp:= Temp+ FparseStr[i];
        if FparseStr[i] = '"' then CloseChar:= '"' else
        if FparseStr[i] = '''' then CloseChar:= '''';
      end;
    end;
  end;
  if Temp<> '' then Add(Temp);
end;

function tSqlParser.PushSpaces: Boolean;
var i, Ind: Integer;
    Temp: tStringList;
    S1: String;
function IsOp(S1: String): Integer;
var j: Integer;
begin
  Result:= 0;
  for j:= 1 to Length(S1) do
    if (S1[j] in Operator) then
    begin
      Result:= j;
      Break;
    end;
end;
begin
  Result:= False;
  Temp:= tStringList.Create;
  for i:= 0 to Count- 1 do
  begin
    if ((Length(Strings[i])= 1) or
        (Pos('"', Strings[i])= 1)) then
    begin
      Temp.Add(Strings[i]);
    end else
    begin
      Ind:= IsOp(Strings[i]);
      if Ind = 0 then Temp.Add(Strings[i]) else
      begin
        S1:= Strings[i];
        Result:= True;
        if Ind = 1 then
        begin
          Temp.Add(Strings[i][1]);
          Temp.Add(Copy(Strings[i], 2, Length(Strings[i])));
        end else
        if Ind = Length(Strings[i])then
        begin
          Temp.Add(Copy(Strings[i], 1, Length(Strings[i])-1));
          Temp.Add(Strings[i][Length(Strings[i])]);
        end else
        begin
          Temp.Add(Copy(Strings[i], 1, Ind- 1));
          Temp.Add(Strings[i][Ind]);
          Temp.Add(Copy(Strings[i], Ind+ 1, Length(Strings[i])));
        end;
      end;
    end;
  end;
  Clear;
  for i:= 0 to Temp.Count-1 do
    Add(Temp.Strings[i]);
  Temp.Free;
end;

procedure tSqlParser.RemoveComments;
var IsCom: Boolean;
    i: Integer;
    Coll: String;
begin
  IsCom:= False;
  for i:= 1 to Length(FParseStr) do
    if IsCom then IsCom:= not(FParseStr[i] = #10) else
      if FParseStr[i] = '#' then IsCom:= True
                            else Coll:= Coll+FParseStr[i];
  FParseStr:= Coll;
end;

procedure tSqlParser.RemoveOuterBrackets;
var i: Integer;
    Temp: tStringList;
begin
  Temp:= tStringList.Create;
  for i:= 0 to Count- 1 do
  begin
    if (Strings[i] = '(') or (Strings[i] = ')') then Continue else
    if (Pos('(', Strings[i])>0) then Strings[i]:= Copy(Strings[i], 2, Length(Strings[i])) else
    if (Pos(')', Strings[i])>0) then Strings[i]:= Copy(Strings[i], 1, Length(Strings[i])-1);
    Temp.Add(Strings[i]);
  end;
  Clear;
  for i:= 0 to Temp.Count- 1 do
    Add(Temp.Strings[i]);
end;

procedure tSqlParser.SetParseStr(aValue: String);
begin
  if FDelimiter = [] then FDelimiter:= [' ',#9, #10, #13];
  if FOperator = [] then FOperator:= ['=', '*', '/', '-', '+', '<', '>', '(', ')', ',', ';', '.', ':'];
  FParseStr:= aValue;
  RemoveComments;       //before parsing
  Parse;
  ConnectStrings;       //because the strings are cutted by delimiters
  while PushSpaces do;  //cutting of items by operators
  if Strings[Count- 1] = ';' then
    Delete(Count- 1);   //last ';' is removed
//  ConnectEnum;          //because the enumerator can contain space (syntax error)
//  RemoveOuterBrackets;
  Upper;                //UpperCase of reserved words
  ConnectOrderBy;       //for simple interpreting
  SplitParse;           //split result to some individual TStringList
end;

procedure tSqlParser.SplitParse;
var i: Integer;
begin
  for i:= 0 to FQueryList.Count- 1 do
  begin
    TStringList(FQueryList.Items[0]).Free;
    FQueryList.Delete(0);
  end;
  if Count= 0 then Exit;
  FQueryList.Add(TStringList.Create);
  for i:= 0 to Count- 1 do
    if Strings[i] <> ';'
    then TStringList(FQueryList.Items[FQueryList.Count- 1]).Add(Strings[i])
    else FQueryList.Add(TStringList.Create);
end;

procedure tSqlParser.Upper;
const ResW = 'ADD,ALTER,AS,ASC,BY,COLUMN,COMMIT,CONNECT,CREATE,DELETE,DESC,DROP,FROM,GROUP,HAVING,INSERT,INTO,LIKE,ORDER,ROLLBACK,SELECT,SET,TABLE,UPDATE,VALUES,WHERE';
var i: Integer;
begin
  for i:= 0 to Count- 1 do
    if Pos(UpperCase(Strings[i]), ResW)<>0 then Strings[i]:= UpperCase(Strings[i]);
end;

constructor tSqlParser.Create;
begin
  FQueryList:= TList.Create;
end;

destructor tSqlParser.Destroy;
begin
  FqueryList.Free;
  inherited;
end;

function tSqlParser.GetQuery(Index: Integer): tStringList;
begin
  if Index < FQueryList.Count
    then Result:= tStringList(FQueryList.Items[Index])
    else Result:= nil;
end;

// TEval

procedure TEval.Setup;
var i: Integer;
begin
  WhatIndex:= -1;
  for i:= 0 to Fields.Count- 1 do
    if UpperCase(Fields.Strings[i])= UpperCase(Expression.Strings[0]) then
    begin
      WhatIndex:= i;
      Break;
    end;
  if WhatIndex = -1 then
    LogFile.RunError:= 'Error: Unknown field name in WHERE clause';
end;

constructor TEval.Create(AOwner: tComponent);
begin
  inherited;
  Expression:= tStringList.Create;
end;

destructor TEval.Destroy;
begin
  Expression.Free;
  inherited;
end;

function TEval.GetResult: Boolean;
var ES0, ES1, ES2, Like: String;
begin
  Result:= False;
  if Recor.Strings[WhatIndex]= '' then Exit;  //value = NULL
  ES0:= Recor.Strings[WhatIndex];
  ES1:= Expression.Strings[1];
  if (Expression.Strings[2][1]= '"') or (Expression.Strings[2][1]= '''')
    then ES2:= Copy(Expression.Strings[2], 2, Length(Expression.Strings[2])-2)
    else ES2:= Expression.Strings[2];
  try
    if ES1 = '>' then
    begin
      if UpperCase(Types.Strings[WhatIndex]) = 'STRING'
      then Result:= ES0 > ES2
      else Result:= StrToInt(ES0) > StrToInt(ES2);
    end else
    if ES1 = '<' then
    begin
      if UpperCase(Types.Strings[WhatIndex]) = 'STRING'
      then Result:= ES0 < ES2
      else Result:= StrToInt(ES0) < StrToInt(ES2);
    end else
    if ES1 = 'LIKE' then    // ES0 LIKE ES2
    begin
      if UpperCase(Types.Strings[WhatIndex]) = 'STRING'
      then begin
        if (ES2[1] = '%')and(ES2[Length(ES2)] = '%')then
        begin
          Like:= Copy(ES2, 2, Length(ES2)- 2);
          Result:= (Pos(Like, ES0) > 0);
        end else
        if (ES2[Length(ES2)] = '%') then
        begin
          Like:= Copy(ES2, 1, Length(ES2)- 1);
          Result:= (Pos(Like, ES0) = 1);
        end else
        if (ES2[1] = '%') then
        begin
          Like:= Copy(ES2, 2, Length(ES2)- 1);
          Result:= (Pos(Like, ES0) > 0) and
                   (Length(ES0) = Pos(Like, ES0)+ Length(ES2)-2);
        end;
      end;
    end else
    if ES1 = '=' then
    begin
      if UpperCase(Types.Strings[WhatIndex]) = 'STRING'
      then Result:= ES0 = ES2
      else Result:= StrToInt(ES0) = StrToInt(ES2);
    end;
  except
    LogFile.RunError:= 'Error: in WHERE clause';
  end;
end;

// TStrGrid

procedure TStrGrid.AnalyseSql;
begin
  if FParser.Strings[0] = 'DROP' then      //DROP TABLE tablename
  begin
    FFileName:= FParser.Strings[2]+ '.xml';
    XmlDrop;
  end else
  if FParser.Strings[0] = 'SELECT' then    //SELECT fname1,fname2 FROM tablename WHERE ...
  begin
    FParser.ConnectEnum;
    FParser.RemoveOuterBrackets;
    FFileName:= GetOriginalName(FParser.Strings[FParser.IndexOf('FROM')+ 1]+ '.xml');
    XmlSelect;
  end else
  if FParser.Strings[0] = 'INSERT' then   //INSERT INTO tablename (fname1, fname2,...) VALUES (value1, value2), ()...
  begin
    FParser.ConnectEnum;
    FParser.RemoveOuterBrackets;
    FFileName:= GetOriginalName(FParser.Strings[2]+ '.xml');
    XmlInsert;
  end else
  if FParser.Strings[0] = 'DELETE' then   //DELETE FROM tablename WHERE fname=value
  begin
    FFileName:= GetOriginalName(FParser.Strings[2]+ '.xml');
    XmlDelete;
  end else
  if FParser.Strings[0] = 'UPDATE' then   //UPDATE tablename SET fname = value WHERE fname=value
  begin
    FFileName:= GetOriginalName(FParser.Strings[1]+ '.xml');
    XmlUpdate;
  end else
  if FParser.Strings[0] = 'CREATE' then   //CREATE TABLE tablename (fname ftype fdefault fother,
  begin                                   //                        fname ftype fdefault fother... )
    FFileName:= FParser.Strings[2]+ '.xml';
    XmlCreate;
  end else
  begin
    LogFile.RunError:= 'Error: Unknown word in SQL expression';
    MakeEmptyDataSet(LogFile.RunError);
  end;
end;

procedure TStrGrid.CopyFromResult(aRes: TStrGrid);
var i: Integer;
begin
  if not Assigned(aRes) then
  begin
    LogFile.RunError:= 'Error: Incoming parameter is NIL';
    MakeEmptyDataset(LogFile.RunError);
  end else
  begin
    for i:= 0 to aRes.Count- 1 do Add(aRes.Strings[i]);
    ColCount:= aRes.ColCount;
    RowCount:= aRes.RowCount;
  end;
end;

function TStrGrid.GetOriginalName(aName: String): String;
var SR: TSearchRec;
begin
  Result:= '';
  Chdir(FDataBase);
  if FindFirst('*.*', faAnyfile, SR) = 0 then
    while FindNext(SR) = 0 do
      if Uppercase(SR.Name)= Uppercase(aName) then
      begin
        Result:= SR.Name;
        Break;
      end;
  FindClose(SR);
end;

procedure TStrGrid.MakeEmptyDataset(aMessage: String);
begin
  Add('Result');
  Add(aMessage);
  Add('');
  RowCount:= 2;
  ColCount:= 1;
end;

function TStrGrid.MakeTableHeader: String;
var i: Integer;
begin
  Result:= '<table fields="';
  for i:= 0 to FFields.Count- 1 do
    Result:= Result+ FFields.Strings[i]+ ',';
  Result:= Copy(Result, 1, Length(Result)- 1);
  Result:= Result+ '" types="';
  for i:= 0 to FTypes.Count- 1 do
    Result:= Result+ FTypes.Strings[i]+ ',';
  Result:= Copy(Result, 1, Length(Result)- 1);
  if Pos('autoinc', FTypes.Text)> 0
    then Result:= Result+ '" lastid="'+ IntToStr(FLastid)+ '">'
    else Result:= Result+ '">';
end;

function TStrGrid.ReadTableData(Line: String): Boolean;    //true if need restructure
var i: integer;
    S1, S2, S3: String;
    AW: tAwkLine;
begin
  Result:= False;
  AW:= tAwkLine.Create(Line, [' ']);
  for i:= 0 to AW.Count- 1 do
    if (Pos('FIELDS', UpperCase(AW.GetString(i)))> 0) then S1:= AW.GetString(i) else
    if (Pos('TYPES', UpperCase(AW.GetString(i)))> 0) then S2:= AW.GetString(i) else
    if (Pos('LASTID', UpperCase(AW.GetString(i)))> 0) then S3:= AW.GetString(i);
  AW.Free;
  S1:= Copy(S1, Pos('"', S1)+1, Length(S1)-Pos('"', S1)- 1);
  AW:= tAwkLine.Create(S1,[',']);
    for i:= 0 to AW.Count- 1 do
    begin
      FFieldU.Add(UpperCase(Aw.GetString(i)));
      FFields.Add(Aw.GetString(i));
    end;
    FCol:= Aw.Count; FRow:= 1;
  AW.Free;
  S1:= Copy(S2, Pos('"', S2)+1, Length(S2)-Pos('"', S2)- 2);
  AW:= tAwkLine.Create(S1,[',']);
    for i:= 0 to AW.Count- 1 do FTypes.Add(Aw.GetString(i));
  AW.Free;
  if S3> '' then
  begin
    S1:= Copy(S3, Pos('"', S3)+1, Length(S3)-Pos('"', S3)- 2);
    FLastid:= StrToInt(S1)+ 1;
    Result:= True;
  end;
end;

procedure TStrGrid.SetOrderFieldType;
var i, dex: Integer;
    S1: String;
begin
  dex:= -100;
  for i:= 0 to FFields.Count- 1 do
    if UpperCase(Strings[OrderIndex]) = UpperCase(FFields.Strings[i]) then
    begin
      dex:= i;
      Break;
    end;
  if dex= -100 then
    LogFile.RunError:='Error: In finding type of order field';
  S1:= UpperCase(FTypes.Strings[dex]);
  if (S1 = 'AUTOINC')or (S1 = 'INTEGER') then OFT:= oftInteger else
  if (S1 = 'BOOLEAN') then OFT:= oftBoolean else
  if (S1 = 'DATETIME') then OFT:= oftDatetime else
  if (S1 = 'FLOAT') then OFT:= oftFloat else
  if (S1 = 'STRING') then OFT:= oftString;
end;

function TStrGrid.StripComas(aStr: String): String;
begin
  Result:= aStr;
  if (Result[1]= '''')or(Result[1]= '"') then Result:= Copy(Result, 2, Length(Result));
  if (Result[Length(Result)]= '''')or(Result[Length(Result)]= '"') then
    Result:= Copy(Result, 1, Length(Result)-1);
end;

procedure TStrGrid.XmlCreate;
var aFile: TextFile;
const PRE = '<?xml version=''1.0'' encoding=''ISO-8859-1''?>';
procedure MakeRecord(SL: tStringList);
var i: Integer;
begin
  if SL.Count= 0 then Exit;
  if SL.Strings[0]= 'PRIMARY' then Exit;
  FFields.Add(SL.Strings[0]);
  for i:= 1 to SL.Count- 1 do
    SL.Strings[i]:= UpperCase(SL.Strings[i]);
  if Pos('AUTO_INCREMENT', SL.Text)>0 then FTypes.Add('autoinc') else
  if Pos('INT', SL.Strings[1])>0  then FTypes.Add('integer') else

  if Pos('DOUBLE', SL.Strings[1])>0  then FTypes.Add('float') else
  if SL.Strings[1] = 'FLOAT' then FTypes.Add('float') else
  if SL.Strings[1] = 'REAL' then FTypes.Add('float') else
  if SL.Strings[1] = 'NUMERIC' then FTypes.Add('float') else
  if SL.Strings[1] = 'DECIMAL' then FTypes.Add('float') else

  if SL.Strings[1] = 'TEXT' then FTypes.Add('string') else
  if SL.Strings[1] = 'VARCHAR' then FTypes.Add('string') else

  if SL.Strings[1] = 'TIMESTAMP' then FTypes.Add('datetime');
end;
procedure GetData;
var Defs: tList;
    i: Integer;
begin
  Defs:= tList.Create;
  for i:= 0 to 3 do FParser.Delete(0);
  Defs.Add(tStringList.Create);
  for i:= 0 to FParser.Count- 1 do
  begin
    if Pos('PRIMARY', UpperCase(FParser.Strings[i]))>0 then Break;
    if FParser.Strings[i] = ','
      then Defs.Add(tStringList.Create)
      else tStringList(Defs.Items[Defs.Count- 1]).Add(FParser.Strings[i]);
  end;
  for i:= 0 to Defs.Count- 1 do
    MakeRecord(tStringList(Defs.Items[i]));
  for i:= 0 to Defs.Count- 1 do tStringList(Defs.Items[i]).Free;
  Defs.Free;
end;
begin
  if FileExists(FDataBase+ FFileName) then
  begin
    LogFile.RunError:= 'Error: Existing table';
    MakeEmptyDataset(LogFile.RunError);
    Exit;
  end;
  try
    AssignFile(aFile, FDataBase+ FFileName); Rewrite(aFile);
    WriteLn(aFile, PRE);
    GetData;
    WriteLn(aFile, MakeTableHeader);
    WriteLn(aFile, '</table>');
    CloseFile(aFile);
  except
    LogFile.RunError:= 'Error: Finding table';
  end;
  if LogFile.RunError='' then MakeEmptyDataset('Command finished succesfully')
                         else MakeEmptyDataset(LogFile.RunError);
  if LogFile.RunError='' then LogFile.Add('Command finished succesfully');
end;

procedure TStrGrid.XmlDelete;
var aFile, aNew: TextFile;
    aLine: String;
    Temp: tStringList;
    Eval: TEval;
const ResW = 'ADD,ALTER,AS,ASC,COLUMN,COMMIT,CONNECT,CREATE,DELETE,DESC,DROP,FROM,GROUP,HAVING,INSERT,INTO,ORDER,ROLLBACK,SELECT,SET,TABLE,UPDATE,VALUES,WHERE';
function GetWhere: tEval;
var Ind, i: Integer;
    S1: String;
begin
  Result:= nil;
  Ind:= FParser.IndexOf('WHERE');
  if Ind > -1 then
  begin
    Result:= tEval.Create(Application);
    for i:= Ind+ 1 to FParser.Count- 1 do
    begin
      S1:= FParser.Strings[i];
      if Pos(S1, ResW)= 0 then Result.Expression.Add(FParser.Strings[i])
                          else Break;
    end;
    Result.Fields:= FFields;
    Result.Types:= FTypes;
    Result.Setup;
  end;
end;
procedure GetTable;   //retrieves table specific data
begin
  ReadTableData(aLine);
  Eval:= GetWhere;
end;
function clWhere: Boolean;   // where clause
begin
  if not Assigned(Eval) then Result:= True else
  begin
    Eval.Recor:= Temp;
    Result:= Eval.GetResult;
  end;
end;
procedure GetRecord;
var S1: String;
    i: Integer;
    CurrentRecord: tStringList;
begin
  Temp:= tStringList.Create;
  CurrentRecord:= tStringList.Create;
  CurrentRecord.Add(aLine);
  ReadLn(aFile, aLine);
  CurrentRecord.Add(aLine);
  for i:= 1 to FCol do     //max FCol items are reading
  begin
    S1:= Copy(aLine, Pos('>', aLine)+1, Length(aLine)- 2);
    S1:= Copy(S1, 1, Pos('<', S1)- 1);
    Temp.Add(S1);
    if Pos('</RECORD', UpperCase(aLine))= 0 then
    begin
      ReadLn(aFile, aLine);
      CurrentRecord.Add(aLine);
    end;
  end;
  if not clWhere then
   for i:= 0 to CurrentRecord.Count- 1 do
     WriteLn(aNew, CurrentRecord.Strings[i]);
  CurrentRecord.Free;
  Temp.Free;
end;
begin
  try
    DeleteFile(FDataBase+ FFileName+ '.bak');
    RenameFile(FDataBase+ FFileName, FDataBase+ FFileName+ '.bak');
    FBakFile:= FFileName+ '.bak';
    AssignFile(aFile, FDataBase+ FBakFile); Reset(aFile);
    AssignFile(aNew, FDataBase+ FFileName); Rewrite(aNew);
    while not EOF(aFile) do
    begin
      ReadLn(aFile, aLine);
      if Pos('<TABLE', UpperCase(aLine))> 0  then
      begin
        GetTable;
        WriteLn(aNew, aLine);
      end else if Pos('<RECORD', UpperCase(aLine))> 0 then GetRecord
      else WriteLn(aNew, aLine);
    end;
    CloseFile(aFile);
    CloseFile(aNew);
  except
    LogFile.RunError:= 'Error: Finding table';
  end;
  if LogFile.RunError='' then MakeEmptyDataset('Command finished succesfully')
                         else MakeEmptyDataset(LogFile.RunError);
  if LogFile.RunError='' then LogFile.Add('Command finished succesfully');
end;

procedure TStrGrid.XmlDrop;
begin
  if FileExists(FDataBase+ FFileName) then
  begin
    DeleteFile(FDataBase+ FFileName+ '.bak');
    RenameFile(FDataBase+ FFileName, FDataBase+ FFileName+ '.bak');
    FBakFile:= FFileName+ '.bak';
    LogFile.Add('Command finished succesfully');
    MakeEmptyDataset('Command finished succesfully');
  end else
  begin
    LogFile.RunError:= 'Error: Table not exists';
    MakeEmptyDataset(LogFile.RunError);
  end;
end;

procedure TStrGrid.XmlInsert;
var aOld, aNew: TextFile;
    aLine: String;
    IndexArr: array[0..255] of Byte;
    AW: tAwkLine;
    Values: tStringList;
procedure GetTable;
var i, j: Integer;
begin
  if ReadTableData(aLine) then aLine:= MakeTableHeader;
  WriteLn(aNew, aLine);
  try
    if FParser.Strings[3] = 'VALUES' then
    begin
      AW:= tAwkLine.Create(FParser.Strings[4], [',']);       //prepare Values
      for i:= 0 to FFields.Count- 1 do
      begin
        Values.Add(StripComas(AW.Strings[i]));
        IndexArr[i]:= i;
      end;
      AW.Free;
    end else
    begin
      AW:= tAwkLine.Create(FParser.Strings[3], [',']);       //prepare IndexArr
      for i:= 0 to FFields.Count- 1 do for j:= 0 to AW.Count- 1 do
        if UpperCase(FFields.Strings[i]) = UpperCase(AW.Strings[j]) then
          IndexArr[j]:= i;
      AW.Free;
      AW:= tAwkLine.Create(FParser.Strings[5], [',']);       //prepare Values
      for i:= 0 to AW.Count- 1 do
        Values.Add(StripComas(AW.Strings[i]));
      AW.Free;
    end;
  except
    LogFile.RunError:= 'Error: Definition of values';
  end;
end;
procedure InsertRecord;
var i: Integer;
function GetValue(Index: Integer): String;
var i: Integer;
begin
  Result:= '';
  for i:= 0 to Values.Count- 1 do
    if IndexArr[i]= Index then Result:= Values.Strings[i];
end;
begin
  WriteLn(aNew, #9+'<record>');
  Write(aNew, #9#9'<field>');
  if UpperCase(FTypes.Strings[0])= 'AUTOINC'
    then Write(aNew, IntToStr(FLastId))
    else Write(aNew, GetValue(0));
  WriteLn(aNew, '</field>');
  for i := 2 to FCol do
  begin
    Write(aNew, #9#9'<field>');
    Write(aNew, GetValue(i-1));
    WriteLn(aNew, '</field>');
  end;
  WriteLn(aNew, #9'</record>');
  WriteLn(aNew, aLine);     //closing table
end;
begin
  Values:= tStringList.Create;
  DeleteFile(FDataBase+ FFileName+ '.bak');
  RenameFile(FDataBase+ FFileName, FDataBase+ FFileName+ '.bak');
  FBakFile:= FFileName+ '.bak';
  AssignFile(aOld, FDataBase+ FBakFile); Reset(aOld);
  AssignFile(aNew, FDataBase+ FFileName); Rewrite(aNew);
  while not EOF(aOld) do
  begin
    ReadLn(aOld, aLine);
    if Pos('<TABLE', UpperCase(aLine))> 0  then GetTable
    else if Pos('</TABLE', UpperCase(aLine))> 0 then InsertRecord
    else WriteLn(aNew, aLine);
  end;
  CloseFile(aOLd);
  CloseFile(aNew);
  Values.Free;
  if UpperCase(FTypes.Strings[0])= 'AUTOINC'
    then MakeEmptyDataset(IntToStr(FLastId))
    else if LogFile.RunError='' then MakeEmptyDataset('Command finished succesfully')
                                else MakeEmptyDataset(LogFile.RunError);

end;

procedure TStrGrid.XmlSelect;
var aFile: TextFile;
    aLine: String;
    NewCol: Integer;
    IndexArr: array[0..255] of Byte;
    Temp: tStringList;
    Eval: TEval;
{$IFDEF DEMO}
    Handled: Integer;
{$ENDIF}
const ResW = 'ADD,ALTER,AS,ASC,COLUMN,COMMIT,CONNECT,CREATE,DELETE,DESC,DROP,FROM,GROUP,HAVING,INSERT,INTO,ORDER,ROLLBACK,SELECT,SET,TABLE,UPDATE,VALUES,WHERE';
function GetWhere: tEval;
var Ind, i: Integer;
    S1: String;
begin
  Result:= nil;
  Ind:= FParser.IndexOf('WHERE');
  if Ind > -1 then
  begin
    Result:= tEval.Create(Application);
    for i:= Ind+ 1 to FParser.Count- 1 do
    begin
      S1:= FParser.Strings[i];
      if Pos(S1, ResW)= 0 then Result.Expression.Add(FParser.Strings[i])
                          else Break;
    end;
    Result.Fields:= FFields;
    Result.Types:= FTypes;
    Result.Setup;
  end;
end;
procedure GetTable;   //retrieves table specific data
var i, j: integer;
    AW: tAwkLine;
begin
  ReadTableData(aLine);
  if FParser.Strings[1] = '*' then
  begin
    NewCol:= FCol;
    for i:= 0 to FFields.Count- 1 do Add(FFields.Strings[i]);
  end else
  begin   //selecting by fieldnames
    try
      AW:= tAwkLine.Create(FParser.Strings[1], [',']);
      NewCol:= AW.Count;
      for i:= 0 to FFields.Count- 1 do for j:= 0 to AW.Count- 1 do
        if UpperCase(FFields.Strings[i]) = UpperCase(AW.Strings[j]) then
          IndexArr[j]:= i;
      for i:= 0 to AW.Count- 1 do
        Add(FFields.Strings[IndexArr[i]]);
      AW.Free;
    except
      LogFile.RunError:= 'Error: in SELECT clause';
    end;
  end;
  Eval:= GetWhere;
end;
function clWhere: Boolean;   // where clause
begin
  if not Assigned(Eval) then Result:= True else
  begin
    Eval.Recor:= Temp;
    Result:= Eval.GetResult;
  end;
end;
procedure GetRecord;
var S1: String;
    i: Integer;
begin
  Temp:= tStringList.Create;
  ReadLn(aFile, aLine);
  for i:= 1 to FCol do     //max FCol items are reading
  begin
    S1:= Copy(aLine, Pos('>', aLine)+1, Length(aLine)- 2);
    S1:= Copy(S1, 1, Pos('<', S1)- 1);
    Temp.Add(S1);
    if Pos('</RECORD', UpperCase(aLine))= 0 then
      ReadLn(aFile, aLine);
  end;
  if clWhere then
  begin
    if FParser.Strings[1] = '*'
      then for i:= 0 to NewCol- 1 do Add(Temp.Strings[i])
      else for i:= 0 to NewCol- 1 do Add(Temp.Strings[IndexArr[i]]);
    Inc(FRow);
  end;
  Temp.Free;
end;
procedure OrderBy;
var i, j, ParInd: Integer;
    LL: tList;
    SL1, SL2: tStringList;
function Compare(Item1, Item2: Pointer): Integer;
begin
  Result:= 0;
  try
    case OFT of
      oftBoolean: Result:= 0;
      oftDatetime:Result:= 0;
      oftFloat:   Result:= 0;
      oftInteger: if StrToInt(tStringList(Item1).Strings[OrderIndex])>
                     StrToInt(tStringList(Item2).Strings[OrderIndex])then Result:= 1 else
                  if StrToInt(tStringList(Item1).Strings[OrderIndex])<
                     StrToInt(tStringList(Item2).Strings[OrderIndex])then Result:= -1 else
                  Result:= 0;
      oftString:  Result:= CompareText(tStringList(Item1).Strings[OrderIndex],     //OI is a global var
                                       tStringList(Item2).Strings[OrderIndex]);
    else LogFile.RunError:= 'Error: Unknown field type in order clause';
    end;
  except
    LogFile.RunError:= 'Error: Data type problem';
  end;
end;
begin
  ParInd:= 0;
  OrderIndex:= -1;
  for i:= 0 to FParser.Count- 1 do
    if FParser.Strings[i]= 'ORDER BY' then
    begin
      ParInd:= i+ 1;
      Break;
    end;
  if not(ParInd < FParser.Count) then
  begin
    LogFile.RunError:= 'Error: Missed param after ORDER BY';
    Exit;
  end;
  for i:= 0 to FCol- 1 do
    if UpperCase(Strings[i])= UpperCase(FParser.Strings[ParInd]) then
    begin
      OrderIndex:= i;
      Break;
    end;
  SetOrderFieldType;
  if OrderIndex < 0 then
  begin
    LogFile.RunError:= 'Error: ORDER BY param is not correct';
    Exit;
  end;
  LL:= tList.Create;
  SL1:= tStringList.Create;         //SL1 contains the selected fieldnames
  for i:= 0 to FCol- 1 do SL1.Add(Strings[i]);
  for i:= 1 to FRow do
  begin
    SL2:= tStringList.Create;
    for j:= 1 to FCol do SL2.Add(GetCells(j-1, i-1));
    LL.Add(SL2);
  end;
  LL.Delete(0);                    //deleting the fieldnames before sorting
  LL.Sort(@Compare);
  Clear;
  for i:= 0 to SL1.Count-1 do Add(SL1.Strings[i]);
  SL1.Free;
  for i:= 0 to LL.Count- 1 do
  begin
    SL2:= LL.Items[i];
    for j:= 0 to SL2.Count- 1 do Add(SL2.Strings[j]);
    SL2.Free;
  end;
  LL.Free;
end;
procedure SettingAs;
var AW: tAwkLine;
    i: Integer;
begin
  AW:= tAwkLine.Create(FParser.Strings[FParser.IndexOf('AS')+ 1], [',']);
  for i:= 0 to AW.Count- 1 do
    Strings[i]:= AW.Strings[i];
  AW.Free;
end;
procedure Distinct;
begin
end;
begin
  try
    {$IFDEF DEMO}
      Handled:= 0;
    {$ENDIF}
    AssignFile(aFile, FDataBase+ FFileName); Reset(aFile);
    while not EOF(aFile) do
    begin
      ReadLn(aFile, aLine);
      if Pos('<TABLE', UpperCase(aLine))> 0 then GetTable else
      if Pos('<RECORD', UpperCase(aLine))> 0 then
      begin
        {$IFDEF DEMO}
          Inc(Handled);
          if Handled > 500 then
            Continue;
        {$ENDIF}
        GetRecord;
      end;
    end;
    CloseFile(aFile);
    FCol:= NewCol;
    if Pos('DISTINCT', FParser.Text)>0 then Distinct;
    if Pos('ORDER BY', FParser.Text)>0 then OrderBy;
    if  FParser.IndexOf('AS')>-1 then SettingAs;
  except
    LogFile.RunError:= 'Error: Unknown word in SQL expression';
  end;
  if LogFile.RunError='' then LogFile.Add('Command finished succesfully')
                         else MakeEmptyDataset(LogFile.RunError);
end;

procedure TStrGrid.XmlUpdate;
var aFile, aNew: TextFile;
    aLine: String;
    Temp: tStringList;
    Eval: TEval;
const ResW = 'ADD,ALTER,AS,ASC,COLUMN,COMMIT,CONNECT,CREATE,DELETE,DESC,DROP,FROM,GROUP,HAVING,INSERT,INTO,ORDER,ROLLBACK,SELECT,SET,TABLE,UPDATE,VALUES,WHERE';
function GetWhere: tEval;
var Ind, i: Integer;
    S1: String;
begin
  Result:= nil;
  Ind:= FParser.IndexOf('WHERE');
  if Ind > -1 then
  begin
    Result:= tEval.Create(Application);
    for i:= Ind+ 1 to FParser.Count- 1 do
    begin
      S1:= FParser.Strings[i];
      if Pos(S1, ResW)= 0 then Result.Expression.Add(FParser.Strings[i])
                          else Break;
    end;
    Result.Fields:= FFields;
    Result.Types:= FTypes;
    Result.Setup;
  end;
end;
procedure GetTable;   //retrieves table specific data
begin
  ReadTableData(aLine);
  Eval:= GetWhere;
end;
function clWhere: Boolean;   // where clause
begin
  if not Assigned(Eval) then Result:= True else
  begin
    Eval.Recor:= Temp;
    Result:= Eval.GetResult;
  end;
end;
procedure WriteTemp;
var i: Integer;
begin
  WriteLn(aNew, #9+'<record>');
  for i := 0 to Temp.Count- 1 do
  begin
    Write(aNew, #9#9'<field>');
    Write(aNew, Temp.Strings[i]);
    WriteLn(aNew, '</field>');
  end;
  WriteLn(aNew, #9'</record>');
end;
procedure GetRecord;
var S1: String;
    i: Integer;
    CurrentRecord: tStringList;
begin
  Temp:= tStringList.Create;
  CurrentRecord:= tStringList.Create;
  CurrentRecord.Add(aLine);
  ReadLn(aFile, aLine);
  CurrentRecord.Add(aLine);
  for i:= 1 to FCol do     //max FCol items are reading
  begin
    S1:= Copy(aLine, Pos('>', aLine)+1, Length(aLine)- 2);
    S1:= Copy(S1, 1, Pos('<', S1)- 1);
    Temp.Add(S1);
    if Pos('</RECORD', UpperCase(aLine))= 0 then
    begin
      ReadLn(aFile, aLine);
      CurrentRecord.Add(aLine);
    end;
  end;
  if clWhere
  then begin
    Temp.Strings[FFieldU.IndexOf(UpperCase(FParser.Strings[3]))]:=
      StripComas(FParser.Strings[5]);
    WriteTemp;
  end
  else for i:= 0 to CurrentRecord.Count- 1 do
    WriteLn(aNew, CurrentRecord.Strings[i]);
  CurrentRecord.Free;
  Temp.Free;
end;
begin
  try
    DeleteFile(FDataBase+ FFileName+ '.bak');
    RenameFile(FDataBase+ FFileName, FDataBase+ FFileName+ '.bak');
    FBakFile:= FFileName+ '.bak';
    AssignFile(aFile, FDataBase+ FBakFile); Reset(aFile);
    AssignFile(aNew, FDataBase+ FFileName); Rewrite(aNew);
    while not EOF(aFile) do
    begin
      ReadLn(aFile, aLine);
      if Pos('<TABLE', UpperCase(aLine))> 0  then
      begin
        GetTable;
        WriteLn(aNew, aLine);
      end else if Pos('<RECORD', UpperCase(aLine))> 0 then GetRecord
      else WriteLn(aNew, aLine);
    end;
    CloseFile(aFile);
    CloseFile(aNew);
  except
    LogFile.RunError:= 'Error: Finding table';
  end;
  if LogFile.RunError='' then MakeEmptyDataset('Command finished succesfully')
                         else MakeEmptyDataset(LogFile.RunError);
  if LogFile.RunError='' then LogFile.Add('Command finished succesfully');
end;

constructor TStrGrid.Create;
begin
  inherited Create;
  FRow := 0; FCol := 0;
  FFields:= tStringList.Create;
  FFieldU:= tStringList.Create;
  FTypes:= tStringList.Create;
  FParser:= tSqlParser.Create;
  FLastId:= -1;
end;

destructor TStrGrid.Destroy;
begin
  FFields.Free;
  FFieldU.Free;
  FTypes.Free;
  FParser.Free;
  inherited;
end;

procedure TStrGrid.SetSql(aValue: String);
var i: Integer;
    SG: TStrGrid;
begin
  Clear;
  FFields.Clear;
  FTypes.Clear;
  FSql:= aValue;
  FParser.Clear;
  FParser.ParseStr:= aValue;
  for i:= 0 to FParser.QueryList.Count- 1 do
  begin
    SG:= TStrGrid.Create;
    SG.DataBase:= FDataBase;
    SG.Parser.Assign(FParser.GetQuery(i));
    SG.AnalyseSql;
    CopyFromResult(SG);
    if (FBakFile = '') and (SG.FBakFile <> '') then
    begin
      FFileName:= SG.FFileName;
      FBakFile:= SG.FBakFile;
    end;
    SG.Free;
  end;
end;

function  TStrGrid.GetCells(ACol, ARow: Integer): string;
begin  Result := Strings[ARow*FCol+ACol]; end;

procedure TStrGrid.SetCells(ACol, ARow: Integer; const Value: string);
begin  Strings[ARow*FCol+ACol] := Value   end;

function  TStrGrid.GetRObject(ARow: Integer): TObject;
begin  Result := Objects[ARow*FCol];  end;

procedure  TStrGrid.PutRObject(ARow: Integer; const AObject: TObject);
begin  Objects[ARow*FCol] := AObject;  end;

procedure  TStrGrid.DeleteRow( Index : Integer );
var i : integer;
begin
  Delete( Index*FCol );
  for i := 2 to FCol do Delete( Index*FCol );
  dec(FRow);
end;

function  TStrGrid.IndexOfObject( AObject : TObject) : Integer;
begin
  Result := inherited IndexOfObject(AObject);
  if Result >= 0 then Result := Result div FCol;
end;

// TXmlSqlDb

function TXmlSqlDb.GetCanUndo: Boolean;
begin
  if FData.FBakFile<> '' then
  begin
    FUndoDataBase:= FDataBase;
    FUndoFileName:= FData.FFileName;
    FUndoBakFile:= FData.FBakFile;
  end;
  Result:= FUndoBakFile <> '';
end;

function TXmlSqlDb.GetErrorStr: String;
begin
  Result:= FLogFile.FError;
end;

function TXmlSqlDb.GetLogText: String;
begin
  Result:= FLogFile.Text;
end;

procedure TXmlSqlDb.InternalOpen;
var i, j, MaxW: integer;
begin
  try
    FLogFile.FError:= '';
    FData := TStrGrid.Create;
    FData.DataBase:= DataBase;
    FData.Sql:= Sql;
    FFldNames.Clear;
    for i := 1 to FData.ColCount do FFldNames.Add( FData[i-1,0] );
    FData.DeleteRow(0);
  except end;
  if FFldNames.Count > 0 then
  begin
    // Calculation Max Field Width and Record Width
    FRecSize := 0;
    for j := 1 to FFldNames.Count do
    begin
      MaxW := Length( FFldNames[j-1] );
      for i := 1 to FData.RowCount do
        if Length(FData[j-1,i-1]) > MaxW then MaxW := Length(FData[j-1,i-1]);
      FFldNames.Objects[j-1] := Pointer(MaxW);
      inc(FRecSize, MaxW);
    end;
    // Initializing field definitions and field creation (TFields)
    InternalInitFieldDefs;
    if DefaultFields then CreateFields;
    BindFields(True);
    // Creating array of offsets for internal use
    // Each [i] element is an offset of field i from starting of record buffer
    GetMem( FFldOffs, sizeof(Integer)*FData.ColCount );
    j := 0;
    for i:=1 to FFldNames.Count do
      begin FFldOffs^[i-1] := j; inc(j, Integer(FFldNames.Objects[i-1]) ); end;
    // Bookmark's creation
    for i := 1 to FData.RowCount do FData.RowObjects[i-1] := Pointer(i);
    FLastBookMark := FData.RowCount;
    FCurRec := -1;
    FRecInfoOfs := FRecSize;
    FRecBufSize := FRecInfoOfs + SizeOf(TRecInfo);
    BookmarkSize := SizeOf( Integer );
  end;
end;
                                                      
procedure TXmlSqlDb.InternalClose;
begin
  if FFldNames.Count > 0 then
  begin
    BindFields(False);
    if DefaultFields then DestroyFields;
    FData.Clear;
    if FData.FBakFile<> '' then
    begin
      FUndoDataBase:= FDataBase;
      FUndoFileName:= FData.FFileName;
      FUndoBakFile:= FData.FBakFile;
    end;
    FData.Free;
    FFldNames.Clear;
    FreeMem( FFldOffs );
    FlastBookMark := 0; FCurRec := -1;
  end;
end;

function TXmlSqlDb.IsCursorOpen: Boolean;
begin
  Result := Assigned(FData) and (FieldCount>0);
end;

procedure TXmlSqlDb.InternalInitFieldDefs;
var i: integer;
begin
  FieldDefs.Clear;
  for i:= 1 to FFldNames.Count do
    TFieldDef.Create( FieldDefs, FFldNames[i-1], ftString,
                      Integer(FFldNames.Objects[i-1]), False, i );
end;

procedure TXmlSqlDb.InternalHandleException;
begin
  Application.HandleException(Self);
end;

{ BookMarks }
procedure TXmlSqlDb.InternalGotoBookmark(Bookmark: Pointer);
var Index: Integer;
begin
  Index := FData.IndexOfObject(TObject(PLongInt(Bookmark)^));
  if Index <> -1 then FCurRec := Index
                 else DatabaseError('Bookmark not found');
end;

procedure TXmlSqlDb.InternalSetToRecord(Buffer: PChar);
begin
  InternalGotoBookmark(@PRecInfo(Buffer + FRecInfoOfs).Bookmark);
end;

function TXmlSqlDb.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag;
end;

procedure TXmlSqlDb.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag := Value;
end;

procedure TXmlSqlDb.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PLongInt(Data)^ := PRecInfo(Buffer + FRecInfoOfs).Bookmark;
end;

procedure TXmlSqlDb.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PRecInfo(Buffer + FRecInfoOfs).Bookmark := PLongInt(Data)^;
end;

function TXmlSqlDb.GetRecordSize: Word;
begin
  Result := FRecSize;
end;

function TXmlSqlDb.AllocRecordBuffer: PChar;
begin
  GetMem(Result, FRecBufSize);
end;

procedure TXmlSqlDb.FreeRecordBuffer(var Buffer: PChar);
begin
  FreeMem(Buffer, FRecBufSize);
end;

function TXmlSqlDb.GetRecord(Buffer: PChar; GetMode: TGetMode;
                    DoCheck: Boolean): TGetResult;
var i : integer;
begin
  if RecordCount < 1 then begin
    InternalInitRecord( ActiveBuffer );
    Result := grEOF;
  end else
  begin
    Result := grOK;
    case GetMode of
      gmNext:    if FCurRec>=RecordCount-1 then Result:=grEOF
                  else Inc(FCurRec);
      gmPrior:   if FCurRec<=0 then Result := grBOF
                  else Dec(FCurRec);
      gmCurrent: if (FCurRec < 0) or (FCurRec >= RecordCount) then
                    Result := grError;
    end;
    if Result = grOK then
    begin
      for i := 1 to FData.ColCount do
        StrLCopy(Buffer+FFldOffs^[i-1], PChar(FData[i-1,FCurRec]),
                 Integer(FFldNames.Objects[i-1]) );
      with PRecInfo(Buffer + FRecInfoOfs)^ do
      begin
        BookmarkFlag := bfCurrent;
        Bookmark := LongInt(FData.RowObjects[FCurRec]);
      end;
    end else
      if (Result = grError) and DoCheck then DatabaseError('No Records');
  end;
end;

procedure TXmlSqlDb.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, RecordSize, 0);
end;

function TXmlSqlDb.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
begin
  StrLCopy(Buffer, ActiveBuffer+FFldOffs^[Field.FieldNo-1], Field.Size);
  Result := PChar(Buffer)^ <> #0;
end;

procedure TXmlSqlDb.SetFieldData(Field: TField; Buffer: Pointer);
begin
  StrLCopy(ActiveBuffer+FFldOffs^[Field.FieldNo-1], Buffer, Field.Size);
  DataEvent(deFieldChange, Longint(Field));
end;

procedure TXmlSqlDb.InternalFirst;
begin
  FCurRec := -1;
end;

procedure TXmlSqlDb.InternalLast;
begin
  FCurRec := RecordCount;
end;

procedure TXmlSqlDb.InternalPost;
begin
end;

procedure TXmlSqlDb.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
end;

procedure TXmlSqlDb.InternalDelete;
begin
end;

function TXmlSqlDb.GetRecordCount: Longint;
begin
  Result := FData.RowCount;
end;

function TXmlSqlDb.GetRecNo: Longint;
begin
  UpdateCursorPos;
  if (FCurRec = -1) and (RecordCount > 0) then
    Result := 1 else
    Result := FCurRec + 1;
end;

procedure TXmlSqlDb.SetRecNo(Value: Integer);
begin
  if (Value >= 0) and (Value < FData.RowCount-1) then
  begin
    FCurRec := Value - 1;
    Resync([]);
  end;
end;

constructor TXmlSqlDb.Create(aOwner: tComponent);
begin
  inherited;
  FFldNames := TStringList.Create;
  FLogFile:= TLogFile.Create;
  FLogFile.Path:= ExtractFilePath(Application.ExeName);
  FLogFile.FileName:= 'xmlsqldb.log';
  FLogFile.MaxLine:= 10;
//  FLogFile.SaveToFile(FLogFile.Path+ FLogFile.FileName);  //reset
  LogFile:= FLogFile;
end;

destructor TXmlSqlDb.Destroy;
begin
  inherited;
  FFldNames.Free;
  LogFile.Free;
end;

procedure TXmlSqlDb.Undo;
begin
  if FUndoBakFile = '' then Exit;
  if FileExists(FUndoDataBase+ FUndoBakFile) then
  begin
    DeleteFile(FUndoDataBase+ FUndoFileName);
    RenameFile(FUndoDataBase+ FUndoBakFile, FUndoDataBase+ FUndoFileName);
  end;
  FUndoBakFile:= '';
  FData.FBakFile:= '';
end;

end.

